/*
 * Copyright (c) KylinSoft  Co., Ltd. 2024. All rights reserved.
 *
 * kaiming-builder is licensed under the GPL v2.0+.
 * 
 * See the LICENSE file for more details.
 */

#ifndef __KMINFOJSON_H__
#define __KMINFOJSON_H__

#include "KMNlohmannJsonHelper.h"

/**
 * @brief : 读写info.json配置文件的类
 */ 
class KMInfoJson
{
public:
    /**
     * 清单内部类，用于描述permission filesystem属性
     */ 
    struct Filesystem
    {
        bool m_xdgUserDirs = false;
        std::vector<std::string> m_filesystemDirs;
    };

    /**
     * 清单内部类，用于描述permission net属性
     */
    struct Net
    {
        bool m_network = false;
        bool m_scanning = false;
    };

   /**
     * 清单内部类，用于描述permission device属性
     */
    struct Device
    {
        bool m_bluetooth = false;
        bool m_microphone = false;
        bool m_camera = false;
        bool m_usb = false;
        bool m_serial = false;
    };

    /**
     * 清单内部类，用于描述permission processManager属性
     */
    struct ProcessManager
    {
        bool m_autostart = false;
    };

    /**
     * 清单内部类，用于描述permission others属性
     */
    struct Others
    {
        bool m_systemNotice = false;
        bool m_systemTime = false;
        bool m_systemServer = false;
        bool m_windowTop = false;
        bool m_preventSleep = false;
        bool m_systemDisplayResolution = false;
        bool m_systemDisplayBrightness = false;
        bool m_systemAudioSetting = false;
        bool m_systemAudioSize = false;
        bool m_systemCrontab = false;
        bool m_systemPowerMode = false;
        bool m_systemPowerPolicy = false;
    };

    /**
     * 清单内部类，用于描述permission属性
     */ 
    struct Permission
    {
        bool m_none = false;
        Filesystem m_filesystem;
        Net m_net;
        Device m_device;
        ProcessManager m_processManager;
        Others m_others;
    };

    /**
     * 扩展字段
     */
    struct Annotations
    {
        // 是否提权。应用如果需要sudo运行，则设置为true
        bool kylin_privilege = false;

        // 是否系统应用，系统应用使用overlay方式挂载并安装在/usr目录下
        bool kylin_sysapp = false;

        // 依赖的app列表
        std::vector<std::string> kylin_depend_apps;
    };

public:
    KMInfoJson() = default;
    ~KMInfoJson() = default;

    /**
     * @brief : 加载json文件
     * @param : [in] jsonFile, json文件
     * @return: nlohmann::json, 解析后的json对象
     * @throw : KMException
     */ 
    void loadFile(const std::string& jsonFile);

    /**
     * @brief : 保存json对象到文件
     * @param : [in] jsonFile, json文件
     * @throw : KMException
     */ 
    void saveFile(const std::string& jsonFile);
    void saveModulesFile(const std::string& jsonFile);

private:
    void fromJson();

    void toJson();

    void determineModule();

    void parsePermissionFromJson();

    void permissionToJson();

    void annotationsToJson();

public:
    // info.json属性域，暂时不进行封装

    // 协议版本号
    std::string schema_version = "1.0";

    // 应用id
    std::string id;

    // 架构
    std::vector<std::string> archs;

    // base
    std::string base;

    // 运行时
    std::string runtime;

    // 渠道
    std::string channel;

    // kaiming run执行命令
    std::vector<std::string> commands;

    // 应用名称
    std::string name;

    // 多语言应用名称，例：name[zh_CN]
    std::map<std::string, std::string> names;

    // 描述
    std::string description;

    // 多语言描述，例：description[zh_CN]
    std::map<std::string, std::string> descriptions;

    // kind {app/runtime/base}
    std::string kind;

    // module (binary、devel)
    std::map<std::string, std::string> modules;
    std::string module;

    // 安装大小
    uint64_t size ;

    // 版本
    std::string version;

    // 环境变量
    std::map<std::string, std::string> envs;

    // 权限
    std::vector<std::string> permissions;
    Permission permission;

    // 许可证信息，SPDX格式
    std::string license;

    // 首页
    std::string homepage;

    // 维护者
    std::vector<std::string> maintainers;

    // 扩展字段
    Annotations annotations;

private:
    // 数据成员域
    nlohmann::json m_root;
};

#endif // !__KMINFOJSON_H__
